# -*- coding: utf-8 -*-
'''
aakbar -- amino-acid k-mer signature tools
'''

# This file makes it easier for developers to test in-place via the command
# python3 -m aakbar
# from the directory above this one.

from .__init__ import cli

if __name__ == '__main__':
    cli(auto_envvar_prefix='AAKBAR')



# -*- coding: utf-8 -*-
'''Simplicity masking and scoring classes.
'''
# 3rd-party packages
from colorama import Fore
import numpy as np
# module packages
from . import cli
from .common import *
#
# global constants
#
TERM_CHAR = '$'
#
# class definitions
#


class RunlengthSimplicity(SimplicityObject):
    '''Define simplicity by the number of repeated letters.

    '''

    def __init__(self, default_cutoff=DEFAULT_SIMPLICITY_CUTOFF):
        super().__init__(default_cutoff=default_cutoff)
        self.label = 'runlength'
        self.desc = 'runlength (repeated characters)'

    def _runlength(self, s):
        return [all([s[i + j + 1] == s[i] for j in range(self.cutoff - 1)])
                for i in range(len(s) - self.cutoff + 1)]

    def mask(self, seq, print_results=False):
        '''Mask high-simplicity positions in a string.

        :param s: Input string.
        :return: Input string with masked positions changed to lower-case.
        '''
        for pos in [i for i, masked in
                    enumerate(self._runlength(to_str(seq).upper()))
                    if masked]:
            if isinstance(seq, str):  # strings need to have whole length set
                seq = seq[:pos] + seq[pos:pos + self.cutoff].lower() + \
                    seq[pos + self.cutoff:]
            else:
                seq[pos:pos +
                    self.cutoff] = to_str(seq[pos:pos +
                                              self.cutoff]).lower()
        return super().mask(seq)


class LetterFrequencySimplicity(SimplicityObject):
    '''Define simplicity by the number of repeated letters.

    '''

    def __init__(self,
                 default_cutoff=DEFAULT_SIMPLICITY_CUTOFF,
                 window_size=None):
        global config_obj
        super().__init__(default_cutoff=default_cutoff)
        if window_size is None:
            try:
                self.window_size = config_obj.config_dict['simplicity_window']
            except KeyError:
                self.window_size = DEFAULT_SIMPLICITY_WINDOW
        else:
            self.window_size = window_size
        self.label = 'letterfreq%d' % self.window_size
        self.desc = 'letter frequency in window of %d residues' % self.window_size


    def mask(self, seq, print_results=False):
        '''Mask high-simplicity positions in a string.

        :param s: Input string.
        :return: Input string with masked positions changed to lower-case.
        '''
        out_str = to_str(seq)
        end_idx = len(out_str) - 1
        byte_arr = np.array([char for char in to_bytes(out_str.upper())])
        mask_positions = set()
        #
        # test character by character for number of occurrances over window
        #
        for char in set(byte_arr):  # test character by character
            char_positions = list(np.where(byte_arr == char)[0])
            while len(char_positions) >= self.cutoff:
                testpos = char_positions.pop(0)
                next_positions = char_positions[:self.cutoff - 1]
                if next_positions[-1] - testpos < self.window_size:
                    mask_positions = mask_positions.union(
                        set([testpos] + next_positions))
        #
        # mask everything
        #
        for pos in mask_positions:
            out_str = out_str[:pos] + out_str[pos].lower() + out_str[pos + 1:]

        if isinstance(seq, str):  # strings need to have whole length set
            seq = out_str
        else:  # may be MutableSeq that needs lengths
            seq[:end_idx] = out_str[:end_idx]
        return super().mask(seq)


class GenerisSimplicity(SimplicityObject):
    '''Define simplicity by the number of repeated letters.

    '''

    def __init__(self,
                 default_cutoff=DEFAULT_SIMPLICITY_CUTOFF,
                 window_size=None):
        super().__init__(default_cutoff=default_cutoff)
        if window_size is None:
            try:
                self.window_size = config_obj.config_dict['simplicity_window']
            except KeyError:
                self.window_size = DEFAULT_SIMPLICITY_WINDOW
        else:
            self.window_size = window_size
        self.label = 'generis%d' % self.window_size
        self.desc = 'pattern by BW xform in window of %d residues' % self.window_size

    def mask_pairs(self, s):
        return [all([s[i + j + 1] == s[i] for j in range(1)])
                for i in range(len(s) - 1)]

    def _bwt(self, s):
        '''Burrows-Wheeler Transform.

        :param s: Input string.  Must not contain TERMCHAR.
        :return: Transformed string.
        '''
        s = s + TERM_CHAR
        return ''.join([x[-1] for x in sorted([s[i:] + s[:i]
                                               for i in range(len(s))])])

    def _ibwt(self, s):
        '''Inverse Burrows-Wheeler Transform on uppercase-only comparisons.

        :param s: Transformed string with mixed upper and lower case.
        :return: Untransformed string with original order.
        '''
        L = [''] * len(s)
        for i in range(len(s)):
            L = sorted([s[i] + L[i] for i in range(len(s))],
                       key=str.upper)
        return [x for x in L if x.endswith(TERM_CHAR)][0][:-1]

    def merge_mask_regions(self, mask, max_run):
        "merge regions separated by less than max_run"
        runs = self.run_lengths(mask)
        for i in range(len(runs)):
            if runs[i] <= max_run:
                mask[i] = True
        return mask

    def unset_small_regions(self, mask, min_run):
        "merge regions separated by less than max_run"
        runs = self.run_lengths([int(not i) for i in mask])
        for i in range(len(runs)):
            if mask[i] and (runs[i] < min_run-1):
                mask[i] = False
        return mask


    def mask(self, seq, print_results=False):
        '''Mask high-simplicity positions in a string.

        :param s: Input string, will be converted to all-uppercase.
        :return: Input string with masked positions changed to lower-case.
        '''
        out_str = to_str(seq)
        end_idx = len(out_str) - 1
        upper_str = out_str.upper()
        # run-length mask in direct space
        dir_mask = self.mask_pairs(upper_str)
        dir_mask = self.merge_mask_regions(dir_mask, 3)
        dir_mask = self.unset_small_regions(dir_mask, self.cutoff)
        for pos in [i for i, masked in
                    enumerate(dir_mask)
                    if masked]:
            out_str = out_str[:pos] + out_str[pos:pos + 2].lower()\
                      + out_str[pos + 2:]
        if print_results:
            logger.info('     rlm: %s', colorize_string(out_str))
        # run-length mask in Burrows-Wheeler space
        bwts = self._bwt(upper_str)
        bwt_mask = self.mask_pairs(bwts)
        bwt_mask = self.merge_mask_regions(bwt_mask, 2)
        bwt_mask = self.unset_small_regions(bwt_mask, self.cutoff)
        for pos in [i for i, masked in
                    enumerate(bwt_mask)
                    if masked]:
            bwts = bwts[:pos] + bwts[pos:pos + \
                self.cutoff].lower() + bwts[pos + self.cutoff:]
        if print_results:
            logger.info('     bwt: %s', colorize_string(bwts))
        ibwts = self._ibwt(bwts)
        if print_results:
            logger.info('    ibwt: %s', colorize_string(ibwts))
        # add in mask from inverse-transformed string
        for pos in [i for i, char in
                    enumerate(ibwts) if char.islower()]:
            out_str = out_str[:pos] + out_str[pos].lower() + out_str[pos + 1:]
        if print_results:
            logger.info(' generis: %s', colorize_string(out_str))
        if isinstance(seq, str):  # strings need to have whole length set
            seq = out_str
        else:                    # may be MutableSeq that needs lengths
            seq[:end_idx] = out_str[:end_idx]
        return  super().mask(seq)

#
# Instantiations of classes.
#
NULL_SIMPLICITY = SimplicityObject()
RUNLENGTH_SIMPLICITY = RunlengthSimplicity()
LETTERFREQ_SIMPLICITY = LetterFrequencySimplicity()
GENERIS_SIMPLICITY = GenerisSimplicity()

@cli.command()
@click.option('--smooth/--no-smooth', default=True, is_flag=True,
              help='Finish with real-space smoothing.')
@click.option('--cutoff', default=DEFAULT_SIMPLICITY_CUTOFF, show_default=True,
              help='Maximum simplicity to keep.')
@click.option('-k', default=DEFAULT_K, show_default=True,
              help='k-mer size for score calculation.')
def demo_simplicity(smooth, cutoff, k):
    '''Demo self-provided simplicity outputs.

    :param cutoff: Simplicity value cutoff, lower is less complex.
    :param window_size: Window size for masking computation..
    :return:
    '''
    user_ctx = get_user_context_obj()
    simplicity_obj = user_ctx['simplicity_object']
    simplicity_obj.set_cutoff(cutoff)
    logger.info('Simplicity function is %s with cutoff of %d.',
                simplicity_obj.desc, cutoff)
    simplicity_obj.set_k(k)
    simplicity_obj.use_smoother(smooth)
    logger.info('           Mask window demo for %d-mers:', k)
    mask_test = 'AAAAAAAAAAaaaaAAAAAAAAAAAaaaaaAAAAAAAAAAAAaaaaaaAAAAAAAAAAAAAaaaaaaaAAAAAAAAAAAAAA'
    logger.info('      in: %s\n S-score: %s\n', mask_test,
                ''.join(['%X' % i for i in
                         simplicity_obj.score(mask_test)]))
    try:
        window_size = user_ctx['simplicity_window']
    except KeyError:
        window_size = DEFAULT_SIMPLICITY_WINDOW
    smoother_tests = [('L end', 'aAAAAAAAAAAAAA'),
                      ('R end', 'AAAAAAAAAAAAAa'),
                      ('singleton', 'AAAAAaAAAAAAAAaaaaAAAAA'),
                      ('non-windowed singleton', 'AaAAAAaAAAAaAAAAAaAAAAAaAAAAAA')]
    logger.info('Demo smoother with window size of %d' %simplicity_obj.k)
    for label, pattern in smoother_tests:
        logger.info('%s:  %s', label, colorize_string(pattern))
        smoothed = simplicity_obj.smoother(pattern)
        logger.info('smoothed: %s', colorize_string(smoothed))
    for desc, case in simplicity_obj.testcases:
        if case is '':
            logger.info('              %s', desc)
        else:
            logger.info('\n%s:', desc)
            logger.info('      in: %s', case)
            masked_str = simplicity_obj.mask(case, print_results=True)
            logger.info('     out: %s', colorize_string(masked_str))


@cli.command()
@click.argument('window_size')
def set_simplicity_window(window_size):
    '''Define size of simplicity window.
    '''
    global config_obj
    if window_size == ():
        try:
            window_size = config_obj.config_dict['simplicity_window']
            default = ''
        except KeyError:
            window_size = DEFAULT_SIMPLICITY_WINDOW
            default = ' (default)'
        logger.info('Window size is %d residues%s',
                    window_size, default)
    try:
        window_size = int(window_size)
    except ValueError:
        logger.error('Window size must be an integer value.')
        sys.exit(1)
    if window_size < 3:
        logger.error('Window size must be >=3.')
        sys.exit(1)
    config_obj.config_dict['simplicity_window'] = window_size
    logger.info(
        'Window size for letter-frequency simiplicity calculation is now %d residues.',
        window_size)
    config_obj.write_config_dict()

@cli.command()
@click.argument('infile', type=str)
@click.argument('outfile', type=str)
def colorize_fasta(infile, outfile):
    "Color lower-case parts of sequence."
    infilepath = Path(infile)
    outfilepath = Path(outfile)
    with infilepath.open(mode='rU') as infh:
        line = infh.readline()
        with outfilepath.open(mode='wt') as outfh:
            for line in infh:
                if line.startswith('>'):
                    outfh.write(line)
                else:
                    outfh.write(colorize_string(line))


# -*- coding: utf-8 -*-
'''Searching for signature occurrances in aakbar.
'''
# standard library imports
import csv
import os
import shutil
from collections import Counter
# external packages
import numpy as np
import pyfaidx
import matplotlib.pyplot as plt
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
# module imports
from . import cli, log_elapsed_time
from .common import *
#
# Global constants
#
RESIDUES_TO_BASES = 3
HISTOGRAM_BINS = 14
#
# Classes begin here.
#


class PeptideSignatureSearcher(object):
    '''Find peptide signatures in sequences.
    '''

    def __init__(self, filestem, sig_frame, k, n_sets, genome_size,
                 nucleotides=False):
        self.filestem = filestem
        self.k = k
        self.sig_frame = sig_frame
        self.n_sets = n_sets
        self.genome_size = genome_size
        self.nucleotide_input = nucleotides
        #
        self.signatures = np.array(
            sig_frame.index,
            dtype=np.dtype(('S%d' % (self.k))))
        self.signatures.sort()
        logger.info('%d %d-mer terms defined in signature file.',
                    len(self.signatures), k)
        # attributes to be initialized per set
        self.input_dict = None
        self.counter = None
        self.code = None
        self.residues_read = None
        self.dir = None
        self.coverage = None
        self.divergence = None
        self.n_seqs = None
        # per-gene attributes
        self.weightarr = None

    def init_set(self, input_dict, code, set_dir):
        global config_obj
        self.input_dict = input_dict
        self.code = code
        self.counter = Counter()
        self.residues_read = 0
        self.n_seqs = 0
        self.coverage = []
        self.divergence = []
        self.sigcountpath = os.path.join(
            set_dir, self.filestem + '_sigcounts.tsv')
        #
        # Signature list initialization
        #
        siglistpath = os.path.join(set_dir,
                                   self.filestem + '_siglist.tsv')
        self.siglistfh = open(siglistpath, 'wt')
        self.siglistwriter = csv.DictWriter(self.siglistfh,
                                            fieldnames=['signature',
                                                        'key',
                                                        'length',
                                                        'position',
                                                        'intersections',
                                                        'max_count',
                                                        'frame'],
                                            delimiter='\t')
        self.siglistwriter.writeheader()
        #
        # Gene list initialization
        #
        genestatspath = os.path.join(set_dir,
                                     self.filestem + '_genestats.tsv')
        self.genestatsfh = open(genestatspath, 'wt')
        self.genestatswriter = csv.DictWriter(self.genestatsfh,
                                              fieldnames=['key',
                                                          'length',
                                                          'coverage',
                                                          'divergence'],
                                              delimiter='\t')
        self.genestatswriter.writeheader()
        #
        # coverage and divergence histograms
        #
        self.coveragehistpath = os.path.join(
            set_dir, self.filestem + '_coveragehist.tsv')
        self.coverageplotpath = os.path.join(
            set_dir, self.filestem + '_coverageplot.' +
            config_obj.config_dict['plot_type'])
        self.divergencehistpath = os.path.join(
            set_dir, self.filestem + '_divergencehist.tsv')
        self.divergenceplotpath = os.path.join(
            set_dir,
            self.filestem +
            '_divergenceplot.' +
            config_obj.config_dict['plot_type'])

    def _count_matches(self, match, terms, key, frame):
        match_positions = np.where(terms == match)[0]
        forwards = bool(frame % 2)
        offset = int(frame / 2)
        if self.nucleotide_input:
            if forwards:
                match_positions = match_positions * 3 + offset
                k = self.k * 3
            else:
                match_positions = len(
                    self.seq) - 1 - match_positions * 3 - offset
                k = self.k * -3
        else:
            k = self.k
        match_str = to_str(match)
        match_count = len(match_positions)
        self.counter[match_str] += match_count
        sig_stats = self.sig_frame.loc[match_str]
        intersections = sig_stats['intersections']
        max_count = sig_stats['max_count']
        for pos in match_positions:
            self.siglistwriter.writerow({
                'signature': to_str(match),
                'key': key,
                'length': len(self.seq),
                'position': pos,
                'intersections': intersections,
                'max_count': max_count,
                'frame': frame})
            for i in range(pos, pos + k):
                self.weightarr[i] = max(self.weightarr[i], intersections)

    def _init_weightarr(self, seq):
        self.seq = seq
        self.weightarr = np.zeros(len(self.seq), dtype=np.int32)

    def _write_weightstats(self, key):
        # calculate per-gene stats
        nonzero = self.weightarr > 0
        coverage = nonzero.astype(int).mean()
        self.coverage.append(coverage)
        if coverage > 0.0:
            divergence = (1. - self.weightarr[nonzero] / self.n_sets).mean()
            self.genestatswriter.writerow({
                'key': key,
                'length': len(self.seq),
                'coverage': coverage,
                'divergence': divergence})
            self.divergence.append(divergence)
        # write footprints
        weight_str = ''.join(['%x' % i for i in self.weightarr])
        if isinstance(self.seq, str):  # strings need to have whole length set
            self.seq = weight_str
        else:  # may be MutableSeq that needs lengths
            end = len(self.seq)
            self.seq[:end] = weight_str[:end]

    def search_as_peptide(self, key):
        s = self.input_dict[key]
        if len(s) == 0:
            logger.warn('  Empty sequence with key "%s".', key)
            return
        self.n_seqs += 1
        self.residues_read += len(s)
        self._init_weightarr(s)
        if self.nucleotide_input:  # do 6-frame translation
            seq_bytes_list = []
            seq = to_bytes(str(s))
            length = len(seq)
            for offset in range(3):  # three bases in a codon
                DNA = Seq(
                    to_str(seq[offset:offset + int((length - offset) / 3) * 3]), generic_dna)
                seq_bytes_list.append(DNA.translate())
                seq_bytes_list.append(DNA.reverse_complement().translate())
        else:
            seq_bytes_list = [to_bytes(str(s))]
        for frame, seq_bytes in enumerate(seq_bytes_list):
            terms = np.array([to_str(seq_bytes[i:i + self.k])
                              for i in
                              range(len(seq_bytes) - self.k)],
                             dtype=np.dtype(('S%d' % (self.k))))
            unique_terms = np.unique(terms)
            for match in np.intersect1d(self.signatures,
                                        unique_terms,
                                        assume_unique=True):
                self._count_matches(match, terms, key, frame)
            self._write_weightstats(key)

    def close_set(self):
        self.siglistfh.close()
        self.genestatsfh.close()
        logger.info('   %d sequences, %d residues read in %s.',
                    self.n_seqs, self.residues_read, self.code)
        if len(self.divergence) == 0:
            logger.warn('No signatures were found in file.')
            return
        if self.genome_size is None:
            self.genome_size = self.residues_read * RESIDUES_TO_BASES
        logger.info('   Mean per-gene coverage is %.2f%%.',
                    np.array(self.coverage).mean() * 100.)
        logger.info('   Genome size for frequency calculations is %d bp.',
                    self.genome_size)
        most_common = self.counter.most_common()
        if len(most_common) > 0:
            top_sig, top_freq = most_common[0]
        else:  # no signatures found
            top_sig = '""'
            top_freq = 0
        logger.info(
            '   Most common signature is %s, which occurs %d times (%e/bp).',
            top_sig,
            top_freq,
            top_freq / self.genome_size)
        signatures = []
        counts = []
        count_freqs = []
        sig_freqs = []
        max_counts = []
        for signature, count in most_common:
            signatures.append(signature)
            counts.append(count)
            count_freqs.append(count / self.genome_size)
            sig_stats = self.sig_frame.loc[signature]
            sig_freqs.append(float(sig_stats['intersections']) / self.n_sets)
            max_counts.append(sig_stats['max_count'])
        #
        # write signature counts
        #
        logger.debug('Writing signature counts file "%s".', self.sigcountpath)
        pd.DataFrame({'counts': counts,
                      'count_freq': count_freqs,
                      'sig_weight': sig_freqs,
                      'max_count': max_counts},
                     columns=['counts',
                              'count_freq',
                              'max_count',
                              'sig_weight'],
                     index=signatures
                     ).to_csv(self.sigcountpath, sep='\t')
        #
        # write and plot coverage histogram
        #
        coverage_hist, bins = np.histogram(np.array(self.coverage) * 100.,
                                           bins=HISTOGRAM_BINS,
                                           range=[0., 100.])
        bin_centers = bins[:-1]  # zero should really be zero
        coverage_hist = coverage_hist * 100. / len(self.coverage)
        logger.debug('Writing coverage histogram to "%s".',
                     self.coveragehistpath)
        pd.Series(coverage_hist,
                  index=bin_centers).to_csv(self.coveragehistpath,
                                            sep='\t',
                                            float_format='%.3f',
                                            header=True)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(bin_centers, coverage_hist)
        plt.title('Coverage Distribution for %s' % (self.code.capitalize()))
        plt.xlabel('Percent Covered')
        plt.ylabel('Percent of Genes')
        plt.savefig(self.coverageplotpath)
        #
        # write divergence histogram
        #
        divergence_hist, bins = np.histogram(self.divergence,
                                             bins=HISTOGRAM_BINS,
                                             range=[0., 1.])

        bin_centers = (bins[:-1] + bins[1:]) / 2.
        divergence_hist = divergence_hist * 100. / len(self.divergence)
        logger.debug('Writing divergence histogram to "%s".',
                     self.divergencehistpath)
        pd.Series(coverage_hist,
                  index=bin_centers).to_csv(self.divergencehistpath,
                                            sep='\t',
                                            float_format='%.3f',
                                            header=True)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(bin_centers, divergence_hist)
        plt.title('Divergence Distribution for %s' % (self.code.capitalize()))
        plt.xlabel('Signature Divergence Score')
        plt.ylabel('Percent of Genes')
        plt.savefig(self.divergenceplotpath)
#
# Cli commands begin here.
#


@cli.command()
@click.option('--genome_size', type=int, default=None,
              help='Genome size in bp for frequency calculations')
@click.option('--nucleotides/--no-nucleotides', default=False,
              help='Input file is nucleotides.')
@click.argument('infilename', type=str)
@click.argument('filestem', type=str)
@click.argument('setlist', nargs=-1, type=DATA_SET_VALIDATOR)
@log_elapsed_time()
def search_peptide_occurrances(
        genome_size,
        nucleotides,
        infilename,
        filestem,
        setlist):
    '''Find signatures in peptide space.

    '''
    global config_obj
    # context inputs
    user_ctx = get_user_context_obj()
    if user_ctx['first_n']:
        logger.info('Only first %d records will be used', user_ctx['first_n'])
    # parameter inputs
    logger.debug('Input file name is "%s".', infilename)
    logger.info('Signature file stem is "%s".', filestem)
    setlist = DATA_SET_VALIDATOR.multiple_or_empty_set(setlist)
    logger.info('Searching in %d data sets:', len(setlist))
    #
    # read signature file
    #
    summarydir = config_obj.config_dict['summary']['dir']
    sigfilepath = os.path.join(summarydir, filestem + '_terms.tsv')
    if not os.path.exists(sigfilepath):
        logger.error('Signature file "%s" does not exist.', sigfilepath)
        sys.exit(1)
    logger.debug('Reading signature file "%s".', sigfilepath)
    sig_frame = pd.read_csv(sigfilepath,
                            usecols=[0, 1, 3],
                            index_col=0,
                            sep='\t')
    k = len(sig_frame.index[0])
    n_sets = max(sig_frame['intersections'])
    outfilestem = os.path.splitext(infilename)[0] + '-' + filestem
    searcher = PeptideSignatureSearcher(outfilestem,
                                        sig_frame,
                                        k,
                                        n_sets,
                                        genome_size,
                                        nucleotides=nucleotides)
    #
    # loop on sets
    #
    for calc_set in setlist:
        set_dir = config_obj.config_dict[calc_set]['dir']
        fastapath = os.path.join(set_dir, infilename)
        if not os.path.exists(fastapath):
            logger.error('Input file "%s" does not exist.', fastapath)
            sys.exit(1)
        footprintpath = os.path.join(set_dir, filestem + '_footprints.faa')
        shutil.copy(fastapath, footprintpath)
        fasta = pyfaidx.Fasta(footprintpath, mutable=True)
        searcher.init_set(fasta, calc_set, set_dir)
        #
        # iterate on genes in FASTA file
        #
        if user_ctx['first_n']:
            keys = list(fasta.keys())[:user_ctx['first_n']]
        else:
            keys = fasta.keys()
        n_recs = len(keys)
        #
        # loop on genes, with or without progress bars
        #
        if user_ctx['progress']:
            with click.progressbar(keys, label='   %s genes processed' % calc_set,
                                   length=n_recs) as bar:
                for key in bar:
                    searcher.search_as_peptide(key)
        else:
            logger.info('  %s: ', calc_set)
            for key in keys:
                searcher.search_as_peptide(key)
        searcher.close_set()
        fasta.close()



# -*- coding: utf-8 -*-
'''Implements shell-level testing'''
from . import cli
# third-party imports
from pytest_shutil import workspace


# Enable the fixture explicitly in your tests or conftest.py (not required when using setuptools entry points)
#pytest_plugins = ['pytest_shutil']

@cli.command()
def test_all():
    # Workspaces contain a handle to the path.py path object (see https://pathpy.readthedocs.io/)
    path = workspace.workspace
    script = path / 'hello.sh'
    script.write_text('#!/bin/sh\n echo hello world!')
    print('hello from test_something!')
    # There is a 'run' method to execute things relative to the workspace root
    workspace.run('hello.sh')


# -*- coding: utf-8 -*-
#
# This file is part of aakbar.
# Copyright (C) 2018, NCGR.
#
# aakbar is free software; you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License; see LICENSE.txt
# file for more details.
#
''' aakbar -- amino-acid k-mer signature tools '''
#
# Developers, install with:
#    pip install -r requirements.txt
#    python setup.py develop
#
from setuptools import setup, find_packages
import sys
# restrict to python 3.4 or later
if sys.version_info < (3,4,0,'final',0):
    raise SystemExit('Python 3.4 or later is required!')

NAME = 'aakbar'

tests_require = [
    'check-manifest>=0.25',
    'coverage>=4.0',
    'isort>=4.2.2.2',
    'pydocstyle>=1.0.0',
    'pytest-cache>=1.0',
    'pytest-cov>=1.8.0',
    'pytest-pep8>=1.0.6',
    'pytest>=2.8.0'
]
extras_require = dict(docs=['Sphinx>=1.4.2'], tests=tests_require)
extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)
packages = find_packages()

setup(
    description=__doc__,
    packages=packages,
    setup_requires=['setuptools>30.3.0',
                    'setuptools-scm>1.5'
                    ],
    install_requires=['Click',],
    entry_points={
        'console_scripts': [NAME + ' = ' + NAME + ':cli']
    },
    use_scm_version={
        'version_scheme': 'guess-next-dev',
        'local_scheme': 'dirty-tag',
        'write_to': NAME + '/version.py'
    },
    extras_require=extras_require,
    tests_require=tests_require,
)


# -*- coding: utf-8 -*-
'''
aakbar -- amino-acid k-mer signature tools.

There are many bioinformatics tools to work on DNA k-mers, but currently none
that are designed to work in amino-acid (protein) space.  aakbar takes as input
FASTA files of called protein genes from two or more genomes and does simplicity
and set logic on the merged lists to create a set of amino-acid peptide signatures.
'''

# standard library imports
import datetime
import functools
import warnings
import logging
import sys
from pathlib import Path
# third-party imports
import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points

# module defs
from .common import get_user_context_obj, logger, config_obj,\
    SimplicityObject, \
    DEFAULT_FILE_LOGLEVEL, DEFAULT_STDERR_LOGLEVEL, AUTHOR, EMAIL, \
    COPYRIGHT, DEFAULT_FIRST_N, VERSION, PROGRAM_NAME, STARTTIME


# private context function
_ctx = click.get_current_context


class CleanInfoFormatter(logging.Formatter):
    def __init__(self, fmt='%(levelname)s: %(message)s'):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        if record.levelno == logging.INFO:
            return record.getMessage()
        return logging.Formatter.format(self, record)


def init_dual_logger(file_log_level=DEFAULT_FILE_LOGLEVEL,
                     stderr_log_level=DEFAULT_STDERR_LOGLEVEL):
    '''Log to stderr and to a log file at different levels
    '''

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            global logger
            # find out the verbose/quiet level
            if _ctx().params['verbose']:
                _log_level = logging.DEBUG
            elif _ctx().params['quiet']:
                _log_level = logging.ERROR
            else:
                _log_level = stderr_log_level
            logger.setLevel(file_log_level)
            stderrHandler = logging.StreamHandler(sys.stderr)
            stderrFormatter = CleanInfoFormatter()
            stderrHandler.setFormatter(stderrFormatter)
            stderrHandler.setLevel(_log_level)
            logger.addHandler(stderrHandler)

            if not _ctx().params['no_log']:  # start a log file
                # If a subcommand was used, log to a file in the
                # logs/ subdirectory of the current working directory
                #  with the subcommand in the file name.
                subcommand = _ctx().invoked_subcommand
                if subcommand is not None:
                    logfile_name = PROGRAM_NAME + '-' + subcommand + '.log'
                    logfile_path = Path('./logs/' + logfile_name)
                    if not logfile_path.parent.is_dir():  # create logs/ dir
                        try:
                            logfile_path.parent.mkdir(mode=0o755, parents=True)
                        except OSError:
                            logger.error(
                                'Unable to create logfile directory "%s"',
                                logfile_path.parent)
                            raise OSError
                    else:
                        if logfile_path.exists():
                            try:
                                logfile_path.unlink()
                            except OSError:
                                logger.error(
                                    'Unable to remove existing logfile "%s"', logfile_path)
                                raise OSError
                    logfileHandler = logging.FileHandler(str(logfile_path))
                    logfileFormatter = logging.Formatter(
                        '%(levelname)s: %(message)s')
                    logfileHandler.setFormatter(logfileFormatter)
                    logfileHandler.setLevel(file_log_level)
                    logger.addHandler(logfileHandler)
            logger.debug('Command line: "%s"', ' '.join(sys.argv))
            logger.debug('%s version %s', PROGRAM_NAME, VERSION)
            logger.debug('Run started at %s', str(STARTTIME)[:-7])

            return f(*args, **kwargs)

        return wrapper

    return decorator


def init_user_context_obj(initial_obj=None):
    '''Put info from global options into user context dictionary
    '''

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            global config_obj
            if initial_obj is None:
                _ctx().obj = {}
            else:
                _ctx().obj = initial_obj
            ctx_dict = _ctx().obj
            if _ctx().params['verbose']:
                ctx_dict['logLevel'] = 'verbose'
            elif _ctx().params['quiet']:
                ctx_dict['logLevel'] = 'quiet'
            else:
                ctx_dict['logLevel'] = 'default'
            for key in ['progress', 'first_n']:
                ctx_dict[key] = _ctx().params[key]
            #
            # simplicity objects
            #
            ctx_dict['simplicity_objects'] = [globals()[key] for key in
                                              globals().keys() if key.endswith('SIMPLICITY')
                                              if isinstance(globals()[key], SimplicityObject)]
            #
            # simplicity objects in plugins
            #
            for entry_point in iter_entry_points('aakbar.simplicity_plugins'):
                ctx_dict['simplicity_objects'].append(entry_point.load())
            # selected simplicity object
            try:
                simplicity_object_label = config_obj.config_dict['simplicity_object_label']
            except KeyError:
                simplicity_object_label = None
            if simplicity_object_label is not None:
                for obj in ctx_dict['simplicity_objects']:
                    if obj.label == simplicity_object_label:
                        ctx_dict['simplicity_object'] = obj
            else:
                try:
                    ctx_dict['simplicity_object'] = ctx_dict['simplicity_objects'][0]
                except IndexError:
                    ctx_dict['simplicity_object'] = None
            return f(*args, **kwargs)

        return wrapper

    return decorator


def log_elapsed_time():
    '''Log the elapsed time
    '''

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            returnobj = f(*args, **kwargs)
            logger.debug(
                'Elapsed time is %s', str(datetime.now() - STARTTIME)[:-7])
            return returnobj

        return wrapper

    return decorator


@with_plugins(iter_entry_points('aakbar.cli_plugins'))
@click.group(epilog=AUTHOR + ' <' + EMAIL + '>.  ' + COPYRIGHT)
@click.option('--warnings_as_errors', '-e', is_flag=True, show_default=True,
              default=False, help='Warnings cause exceptions.')
@click.option('-v', '--verbose', is_flag=True, show_default=True,
              default=False, help='Log debugging info to stderr.')
@click.option('-q', '--quiet', is_flag=True, show_default=True,
              default=False, help='Suppress logging to stderr.')
@click.option('--no_log', is_flag=True, show_default=True,
              default=False, help='Suppress logging to file.')
@click.option('--progress', is_flag=True, show_default=True,
              default=False, help='Show a progress bar, if supported.')
@click.option('--first_n', default=DEFAULT_FIRST_N,
              help='Process only this many records. [default: all]')
@click.version_option(version=VERSION, prog_name=PROGRAM_NAME)
@init_dual_logger()
@init_user_context_obj()
def cli(warnings_as_errors, verbose, quiet,
        progress, first_n, no_log):
    """aakbar -- amino-acid k-mer signature tools

    If COMMAND is present, and --no_log was not invoked,
    a log file named akbar-COMMAND.log
    will be written in the ./logs/ directory.
    """
    if warnings_as_errors:
        logger.debug(
            'Runtime warnings (e.g., from pandas) will cause exceptions')
        warnings.filterwarnings('error')


@cli.command()
@log_elapsed_time()
def test_logging():
    '''Log at different severity levels.

        Example:
            aakbar test_logging
    '''
    logger.debug('debug message')
    logger.info('info message')
    logger.warning('warning message')
    logger.error('error message')


@cli.command()
def show_context_object():
    '''Print the global context object.
    '''
    user_ctx = get_user_context_obj()
    logger.info('User context object')
    for key in user_ctx.keys():
        logger.info('   %s: %s', key, user_ctx[key])


# import other functions
from .config import show_config, define_set, label_set,\
    define_summary, init_config_file, set_simplicity_type
from .core import calculate_peptide_terms, filter_peptide_terms,\
    intersect_peptide_terms, peptide_simplicity_mask, install_demo_scripts
from .simplicity import set_simplicity_window, demo_simplicity, colorize_fasta, \
    NULL_SIMPLICITY, RUNLENGTH_SIMPLICITY, LETTERFREQ_SIMPLICITY, GENERIS_SIMPLICITY
from .search import search_peptide_occurrances
from .plot import conserved_signature_stats
from .test import test_all

SIMPLICITY_OBJECT = SimplicityObject
